/*
 * wrjpgcom.c
 *
 * Copyright (C) 1994-1997, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains a very simple stand-alone application that inserts
 * user-supplied text as a COM (comment) marker in a JFIF file.
 * This may be useful as an example of the minimum logic needed to parse
 * JPEG markers.
 */

#include <stdio.h>
#include <assert.h>

#ifdef USE_SETMODE
#include <fcntl.h>		/* to declare setmode()'s parameter macros */
/* If you have setmode() but not <io.h>, just delete this line: */
#include <io.h>			/* to declare setmode() */
#endif

#ifdef DONT_USE_B_MODE		/* define mode parameters for fopen() */
#	define READ_BINARY	"r"
#	define WRITE_BINARY	"w"
#else
#	ifdef VMS			/* VMS is very nonstandard */
#		define READ_BINARY	"rb", "ctx=stm"
#		define WRITE_BINARY	"wb", "ctx=stm"
#	else				/* standard ANSI-compliant case */
#		define READ_BINARY	"rb"
#		define WRITE_BINARY	"wb"
#	endif
#endif

/* Reduce this value if your malloc() can't allocate blocks up to 64K.
 * On DOS, compiling in large model is usually a better solution.
 */

#ifndef MAX_COM_LENGTH
#define MAX_COM_LENGTH 65000L	/* must be <= 65533 in any case */
#endif

typedef struct context_s {
	FILE* input;
	FILE* output;
} *context;

#define progname "jpegcomment"

/* Error exit handler */
#define WARN(msg)  (fprintf(stderr, "%s: %s\n", progname, msg))

/* Read one byte, testing for EOF */
static inline int
read_1_byte (context ctxt)
{
	int c;

	c = getc(ctxt->input);
	if (c == EOF)
		WARN("Premature EOF in JPEG file");
	return c;
}

/* Read 2 bytes, convert to unsigned int */
/* All 2-byte quantities in JPEG markers are MSB first */
static inline unsigned int
read_2_bytes (context ctxt)
{
	int c1, c2;

	c1 = read_1_byte(ctxt);
	c2 = read_1_byte(ctxt);
	return (((unsigned int) c1) << 8) + ((unsigned int) c2);
}


/* Routines to write data to output file */

static inline void
write_1_byte (context ctxt,int c)
{
	putc(c,ctxt->output);
}

static inline void
write_2_bytes (context ctxt,unsigned int val)
{
	write_1_byte(ctxt,(val >> 8) & 0xFF);
	write_1_byte(ctxt,val & 0xFF);
}

#define BLOCKSIZE 4096

static void
write_variable_bytes(context ctxt, const char* buffer, size_t length) {
	write_2_bytes(ctxt, length + 2);
	while(length > BLOCKSIZE) {
		size_t written = fwrite(buffer,1,BLOCKSIZE,ctxt->output);
		length -= written;
		buffer += written;
	}
	if(length)
		fwrite(buffer,1,length,ctxt->output);
}

static void
write_marker (context ctxt, int marker)
{
	write_1_byte(ctxt,0xFF);
	write_1_byte(ctxt,marker);
}

static void
copy_rest_of_file (context ctxt)
{
	char buffer[BLOCKSIZE];
	while(!(feof(ctxt->input) || feof(ctxt->output))) {
		size_t readed = fread(buffer,1,BLOCKSIZE,ctxt->input);
		size_t writed = fwrite(buffer,1,readed,ctxt->output);
		assert(readed==writed);
	}
}


/*
 * JPEG markers consist of one or more 0xFF bytes, followed by a marker
 * code byte (which is not an FF).  Here are the marker codes of interest
 * in this program.  (See jdmarker.c for a more complete list.)
 */

#define M_SOF0  0xC0		/* Start Of Frame N */
#define M_SOF1  0xC1		/* N indicates which compression process */
#define M_SOF2  0xC2		/* Only SOF0-SOF2 are now in common use */
#define M_SOF3  0xC3
#define M_SOF5  0xC5		/* NB: codes C4 and CC are NOT SOF markers */
#define M_SOF6  0xC6
#define M_SOF7  0xC7
#define M_SOF9  0xC9
#define M_SOF10 0xCA
#define M_SOF11 0xCB
#define M_SOF13 0xCD
#define M_SOF14 0xCE
#define M_SOF15 0xCF
#define M_SOI   0xD8		/* Start Of Image (beginning of datastream) */
#define M_EOI   0xD9		/* End Of Image (end of datastream) */
#define M_SOS   0xDA		/* Start Of Scan (begins compressed data) */
#define M_COM   0xFE		/* COMment */


/*
 * Find the next JPEG marker and return its marker code.
 * We expect at least one FF byte, possibly more if the compressor used FFs
 * to pad the file.  (Padding FFs will NOT be replicated in the output file.)
 * There could also be non-FF garbage between markers.  The treatment of such
 * garbage is unspecified; we choose to skip over it but emit a warning msg.
 * NB: this routine must not be used after seeing SOS marker, since it will
 * not deal correctly with FF/00 sequences in the compressed image data...
 */

static int
next_marker (context ctxt)
{
	int c;
	int discarded_bytes = 0;

	/* Find 0xFF byte; count and skip any non-FFs. */
	c = read_1_byte(ctxt);
	while (c != 0xFF) {
		discarded_bytes++;
		c = read_1_byte(ctxt);
	}
	/* Get marker code byte, swallowing any duplicate FF bytes.  Extra FFs
	 * are legal as pad bytes, so don't count them in discarded_bytes.
	 */
	do {
		c = read_1_byte(ctxt);
	} while (c == 0xFF);

	if (discarded_bytes != 0) {
		fprintf(stderr, "Warning: garbage data found in JPEG file\n");
	}

	return c;
}


/*
 * Read the initial marker, which should be SOI.
 * For a JFIF file, the first two bytes of the file should be literally
 * 0xFF M_SOI.  To be more general, we could use next_marker, but if the
 * input file weren't actually JPEG at all, next_marker might read the whole
 * file and then return a misleading error message...
 */

static int
first_marker (context ctxt)
{
	int c1, c2;

	c1 = read_1_byte(ctxt);
	c2 = read_1_byte(ctxt);
	if (c1 != 0xFF || c2 != M_SOI)
		WARN("Not a JPEG file");
	return c2;
}


/*
 * Most types of marker are followed by a variable-length parameter segment.
 * This routine skips over the parameters for any marker we don't otherwise
 * want to process.
 * Note that we MUST skip the parameter segment explicitly in order not to
 * be fooled by 0xFF bytes that might appear within the parameter segment;
 * such bytes do NOT introduce new markers.
 */

static void
copy_variable (context ctxt)
/* Copy an unknown or uninteresting variable-length marker */
{
	unsigned int length;
	char buffer[BLOCKSIZE];
	size_t readen,written;

	/* Get the marker parameter length count */
	length = read_2_bytes(ctxt);
	write_2_bytes(ctxt,length);
	/* Length includes itself, so must be at least 2 */
	if (length < 2) {
		WARN("Erroneous JPEG marker length");
		return;
	}
	length -= 2;

	/* Copy the rest of the blob */

	while(length > BLOCKSIZE) {
		readen = fread(buffer,1,BLOCKSIZE,ctxt->input);
	 	written = fwrite(buffer,1,readen,ctxt->output);
		assert(readen==written);
		length -= written;
	}
	if(length>0) {
		readen = fread(buffer,1,length,ctxt->input);
		written = fwrite(buffer,1,length,ctxt->output);
		assert(readen==written);
	}
}

static void
skip_variable (context ctxt)
/* Skip over an unknown or uninteresting variable-length marker */
{
	unsigned int length;

	/* Get the marker parameter length count */
	length = read_2_bytes(ctxt);
	/* Length includes itself, so must be at least 2 */
	if (length < 2) {
		WARN("Erroneous JPEG marker length");
		return;
	}
	length -= 2;

	/* Skip over the remaining bytes */
	if(fseek(ctxt->input,length,SEEK_CUR) < 0) {
		/* grumble, seeking manually, grumble */
		char buffer[BLOCKSIZE];
		while (length > 0)
			length -= fread(buffer,1,BLOCKSIZE,ctxt->input);
	}
}

/*
 * Parse the marker stream until SOFn or EOI is seen;
 * copy data to output, but discard COM markers unless keep_COM is true.
 */

static int
scan_JPEG_header (context ctxt, int keep_COM)
{
	int marker;

	/* Expect SOI at start of file */
	if (first_marker(ctxt) != M_SOI)
		WARN("Expected SOI marker first");
	write_marker(ctxt,M_SOI);

	/* Scan miscellaneous markers until we reach SOFn. */
	for (;;) {
		marker = next_marker(ctxt);
		switch (marker) {
			/* Note that marker codes 0xC4, 0xC8, 0xCC are not, and must not be,
			 * treated as SOFn.  C4 in particular is actually DHT.
			 */
		case M_SOF0:		/* Baseline */
		case M_SOF1:		/* Extended sequential, Huffman */
		case M_SOF2:		/* Progressive, Huffman */
		case M_SOF3:		/* Lossless, Huffman */
		case M_SOF5:		/* Differential sequential, Huffman */
		case M_SOF6:		/* Differential progressive, Huffman */
		case M_SOF7:		/* Differential lossless, Huffman */
		case M_SOF9:		/* Extended sequential, arithmetic */
		case M_SOF10:		/* Progressive, arithmetic */
		case M_SOF11:		/* Lossless, arithmetic */
		case M_SOF13:		/* Differential sequential, arithmetic */
		case M_SOF14:		/* Differential progressive, arithmetic */
		case M_SOF15:		/* Differential lossless, arithmetic */
			return marker;

		case M_SOS:			/* should not see compressed data before SOF */
			WARN("SOS without prior SOFn");
			break;

		case M_EOI:			/* in case it's a tables-only JPEG stream */
			return marker;

		case M_COM:			/* Existing COM: conditionally discard */
			if (keep_COM) {
				write_marker(ctxt,marker);
				copy_variable(ctxt);
			} else {
				skip_variable(ctxt);
			}
			break;

		default:			/* Anything else just gets copied */
			write_marker(ctxt,marker);
			/* we assume it has a parameter count... */
			copy_variable(ctxt);
			break;
		}
	} /* end loop */
}

/* Add a comment from an old to a new jpeg file */

int addComment(
	int inhandle, int outhandle,
	const char* comment, unsigned int comment_length,
	unsigned char keep_COM) {

	int marker;
	struct context_s ctxt;

	inhandle = dup(inhandle);
	outhandle = dup(outhandle);

	printf("Comment is: %s\n",comment);

	if ((ctxt.input = fdopen(inhandle, READ_BINARY)) == NULL) {
		fprintf(stderr, "%s: can't reopen input\n", progname);
		return 0;
	}
	if ((ctxt.output = fdopen(outhandle, WRITE_BINARY)) == NULL) {
		fprintf(stderr, "%s: can't reopen output\n", progname);
		return 0;
	}


	/* Copy JPEG headers until SOFn marker;
	 * we will insert the new comment marker just before SOFn.
	 * This (a) causes the new comment to appear after, rather than before,
	 * existing comments; and (b) ensures that comments come after any JFIF
	 * or JFXX markers, as required by the JFIF specification.
	 */
	marker = scan_JPEG_header(&ctxt,keep_COM);
	/* Insert the new COM marker, but only if nonempty text has been supplied */
	if (comment_length > 0) {
		write_marker(&ctxt,M_COM);
		write_variable_bytes(&ctxt,comment,comment_length);	
	}

	/* Duplicate the remainder of the source file.
	 * Note that any COM markers occuring after SOF will not be touched.
	 */
	write_marker(&ctxt,marker);
	copy_rest_of_file(&ctxt);

	fclose(ctxt.input);
	fclose(ctxt.output);
	return 1;
}
